# -*- coding: utf-8 -*-

from __future__ import absolute_import
from aiida.orm import CalculationFactory, DataFactory, Code
from aiida.orm.data.base import Float
from aiida.work import WorkChain, workfunction, submit
from aiida.work.workchain import ToContext, Outputs
from copy import deepcopy
from aiida_raspa.workflows import RaspaConvergeWorkChain
from math import sqrt

CifData = DataFactory("cif")
ParameterData = DataFactory("parameter")
SingleFile = DataFactory("singlefile")
SinglefileData = DataFactory("singlefile")
RaspaCalculation = CalculationFactory("raspa")
ZeoppCalculation = CalculationFactory("zeopp.network")
NetworkParameters = DataFactory("zeopp.parameters")


@workfunction
def update_raspa_parameters(parameters, pressure):
    """Store input parameters of Raspa for given pressure.

    Note: In order to keep the provenance of both Raspa calculations,
    changing the pressure force us to create a new ParameterData node.
    "workfunctions" will take care of linking the user-provided ParameterData
    node to the new one containing the pressure.
    """
    param_dict = parameters.get_dict()
    param_dict["GeneralSettings"]["ExternalPressure"] = pressure.value
    return ParameterData(dict=param_dict)


def dict_merge(dct, merge_dct):
    """ Taken from https://gist.github.com/angstwad/bf22d1822c38a92ec0a9
    Recursive dict merge. Inspired by :meth:``dict.update()``, instead of
    updating only top-level keys, dict_merge recurses down into dicts nested
    to an arbitrary depth, updating keys. The ``merge_dct`` is merged into
    ``dct``.
    :param dct: dict onto which the merge is executed
    :param merge_dct: dct merged into dct
    :return: None
    """
    import collections

    for k, v in merge_dct.iteritems():
        # This method brakes because component is a list that contains a dict.
        if k == "Component":
            component_dict = dct["Component"][0]
            for k, v in merge_dct["Component"][0].iteritems():
                component_dict[k] = v
            merge_dct["Component"] = [component_dict]
        elif (
            k in dct
            and isinstance(dct[k], dict)
            and isinstance(merge_dct[k], collections.Mapping)
        ):
            dict_merge(dct[k], merge_dct[k])

        else:
            dct[k] = merge_dct[k]


@workfunction
def merge_ParameterData(p1, p2):
    p1_dict = p1.get_dict()
    p2_dict = p2.get_dict()
    dict_merge(p1_dict, p2_dict)
    return ParameterData(dict=p1_dict).store()


class TailCorrectionCheckBlockedPocketsHenry(WorkChain):
    """Test the magic of tailcorrections"""

    @classmethod
    def define(cls, spec):
        """Define workflow specification.

        This is the most important method of a Workchain, which defines the
        inputs it takes, the logic of the execution and the outputs
        that are generated in the process.
        """
        super(TailCorrectionCheckBlockedPocketsHenry, cls).define(spec)

        # zeopp
        spec.input("zeopp_code", valid_type=Code)
        spec.input("_zeopp_options", valid_type=dict, default=None, required=False)
        spec.input("zeopp_probe_radius", valid_type=Float)
        spec.input(
            "zeopp_atomic_radii",
            valid_type=SinglefileData,
            default=None,
            required=False,
        )

        # First we define the inputs, specifying the type we expect
        spec.input("structure", valid_type=CifData, required=True)
        spec.input("raspa_parameters", valid_type=ParameterData, required=True)
        spec.input("raspa_code", valid_type=Code, required=True)

        # The outline describes the basic logic that defines
        # which steps are executed in what order and based on
        # what conditions. Each `cls.method` is implemented below
        spec.outline(
            cls.init,
            cls.run_zeopp,
            cls.run_raspa_widom,  # for henry coefficient
            cls.extract_results,  # create dictionary with results
        )

        # Here we define the output the Workchain will generate and
        # return. Dynamic output allows a variety of AiiDA data nodes
        # to be returned
        spec.dynamic_output()

        # Options are used to retrieve RDF if needed

    def init(self):
        """Initialize variables."""

        self.ctx.options = {
            "resources": {
                "num_machines": 1,  # run on 1 node
                "tot_num_mpiprocs": 1,  # use 1 process
                "num_mpiprocs_per_machine": 1,
            },
            "max_wallclock_seconds": 72 * 60 * 60,  # 72 h walltime
            "max_memory_kb": 2000000,  # 16GB memory
            "queue_name": "serial",  # slurm serial partition
            "withmpi": False,  # we run in serial mode
        }

        self.ctx.number_mc_cycles = 60000
        self.ctx.print_every = int(self.ctx.number_mc_cycles / 10)
        self.ctx.number_init_cylces = 15000

    def run_zeopp(self):
        """Main function that performs zeo++ VOLPO and block calculations."""
        params = {
            "ha": True,
            # 100 samples / Ang^3: accurate for all the structures
            "block": [self.inputs.zeopp_probe_radius.value, 100],
        }

        inputs = {
            "code": self.inputs.zeopp_code,
            "structure": self.inputs.structure,
            "parameters": NetworkParameters(dict=params).store(),
            "_options": self.inputs._zeopp_options,
            "_label": "ZeoppVolpoBlock",
        }

        # Use default zeopp atomic radii only if a .rad file is not specified
        try:
            inputs["atomic_radii"] = self.inputs.zeopp_atomic_radii
            self.report("Zeopp will use atomic radii from the .rad file")
        except:
            self.report("Zeopp will use default atomic radii")

        # Create the calculation process and launch it
        running = submit(ZeoppCalculation.process(), **inputs)
        self.report(
            "pk: {} | Running zeo++ volpo and block calculations".format(running.pid)
        )
        return ToContext(zeopp=Outputs(running))

    def run_raspa_widom(self):
        """Run a Widom calculation in Raspa to get the Henry coefficient"""
        raspa_widom_default_param = {
            "GeneralSettings": {
                "NumberOfInitializationCycles": 0,
                "NumberOfCycles": self.ctx.number_mc_cycles,
                "PrintPropertiesEvery": self.ctx.print_every,  # info on henry coeff
                "PrintEvery": self.ctx.print_every,  # info on loading
            },
            "Component": [{"WidomProbability": 1.0}],
        }
        self.ctx.raspa_widom_parameters = deepcopy(raspa_widom_default_param)

        bs = self.ctx.zeopp["block"]
        with open(bs.get_abs_path() + "/path/" + bs.get_folder_list()[0]) as f:
            self.ctx.number_blocking_spheres = int(f.readline().strip())

        dict_merge(
            self.ctx.raspa_widom_parameters, self.inputs.raspa_parameters.get_dict()
        )

        # Create the input dictionary
        inputs = {
            "code": self.inputs.raspa_code,
            "structure": self.inputs.structure,
            "parameters": ParameterData(dict=self.ctx.raspa_widom_parameters),
            "_options": self.ctx.options,
            "settings": ParameterData(
                dict={
                    "additional_retrieve_list": [
                        "RadialDistributionFunctions/System_0/*"
                    ]
                }
            ),
        }

        if self.ctx.number_blocking_spheres > 0:
            inputs["block_component_0"] = self.ctx.zeopp["block"]
            self.report(
                "Blocking spheres ({}) are present and used for Raspa".format(
                    self.ctx.number_blocking_spheres
                )
            )
        else:
            self.report("No blocking spheres found")

        # Create the calculation process and launch it
        running = submit(RaspaConvergeWorkChain, **inputs)
        self.report(
            "pk: {} | Running raspa for the Henry coefficients".format(running.pid)
        )

        return ToContext(raspa_widom=Outputs(running))

    def extract_results(self):
        """Extract results of the workflow.

        Attaches the results of the raspa calculation and the initial structure to the outputs.
        """

        result = {
            # widom calculation
            "henry_coefficient_widom_average": self.ctx.raspa_widom[
                "component_0"
            ].dict.henry_coefficient_average,
            "henry_coefficient_widom_dev": self.ctx.raspa_widom[
                "component_0"
            ].dict.henry_coefficient_dev,
            "henry_coefficient_widom_units": self.ctx.raspa_widom[
                "component_0"
            ].dict.henry_coefficient_units,
            "adsorption_energy_widom_average": self.ctx.raspa_widom[
                "component_0"
            ].dict.adsorption_energy_widom_average,
            "adsorption_energy_widom_dev": self.ctx.raspa_widom[
                "component_0"
            ].dict.adsorption_energy_widom_dev,
            "adsorption_energy_widom_units": self.ctx.raspa_widom[
                "component_0"
            ].dict.adsorption_energy_widom_units,
            "chemical_potential_average": self.ctx.raspa_widom[
                "component_0"
            ].dict.chemical_potential_average,
            "chemical_potential_dev": self.ctx.raspa_widom[
                "component_0"
            ].dict.chemical_potential_dev,
            "widom_host_ads_total_energy_average": self.ctx.raspa_widom[
                "output_parameters"
            ].get_attr("host_ads_total_energy_average"),
            "widom_host_ads_total_energy_dev": self.ctx.raspa_widom[
                "output_parameters"
            ].get_attr("host_ads_total_energy_dev"),
            "widom_host_ads_total_energy_units": self.ctx.raspa_widom[
                "output_parameters"
            ].get_attr("host_ads_total_energy_unit"),
            "widom_host_ads_vdw_energy_average": self.ctx.raspa_widom[
                "output_parameters"
            ].get_attr("host_ads_vdw_energy_average"),
            "widom_host_ads_vdw_energy_dev": self.ctx.raspa_widom[
                "output_parameters"
            ].get_attr("host_ads_vdw_energy_dev"),
            "widom_total_energy_average": self.ctx.raspa_widom[
                "output_parameters"
            ].get_attr("total_energy_average"),
            "widom_total_energy_dev": self.ctx.raspa_widom[
                "output_parameters"
            ].get_attr("total_energy_dev"),
            "widom_total_energy_units": self.ctx.raspa_widom[
                "output_parameters"
            ].get_attr("total_energy_units"),
            "widom_tail_correction_energy_average": self.ctx.raspa_widom[
                "output_parameters"
            ].get_attr("tail_correction_energy_average"),
            "widom_tail_correction_energy_dev": self.ctx.raspa_widom[
                "output_parameters"
            ].get_attr("tail_correction_energy_dev"),
            "widom_tail_correction_energy_units": self.ctx.raspa_widom[
                "output_parameters"
            ].get_attr("tail_correction_energy_units"),
        }

        self.out("result", ParameterData(dict=result).store())

        self.report("Workchain <{}> completed successfully".format(self.calc.pk))
        return
